/* SPDX-License-Identifier: GPL-2.0-only */
/*
 * Copyright 2023 Red Hat
 */

#ifndef UDS_LINUX_BITS_H
#define UDS_LINUX_BITS_H

/*
 * User-space header providing necessary define provided by the same kernel-space header.
 */

#define BITS_PER_BYTE 8

#endif /* UDS_LINUX_BITS_H */
